@@ -27,8 +27,9 @@
 package as_cui_matrix;
 
 use strict;
-use Curses;         # OpenPKG perl-curses
-use Curses::UI;     # OpenPKG perl-curses
+use Curses;             # OpenPKG perl-curses
+use Curses::UI;         # OpenPKG perl-curses
+use Curses::UI::Common; # OpenPKG perl-curses
 
 my $matrixdata = {};
 my $verbose  = 0;
@@ -39,7 +40,8 @@ my $vrscroll = 0; #callback tells inner loop to scroll wholly view n steps verti
 my $datacoll = 0; #callback tells inner loop to redraw wholly view while trying to reuse widgets if size permits
 my $vcpos    = 0; #callback tells inner loop to scroll wholly view horizontally (neg = left, pos = right)
 my $vrpos    = 0; #callback tells inner loop to scroll wholly view vertically (neg = up, pos = down)
-my $editcell = 0;
+my $editcell = 0; #flags edit or movement mode
+my $escapeit = 0; #flags users desire to escape from popups and cancel the action
 my $widget = {};
 my $matrixwidget = {};
 my $matrixwidgetvisualcell = {};
@@ -466,17 +468,38 @@ sub layoutmatrixwidget {
     &updatecolheading();
     &updaterowheading();
 
+    #   underlines on primary sort key
+    my $ulc = $matrixdata->{"Sortorder"}->[0];
+    $dr = $matrixwidget->{"FirstVisualRow"};
+    my $data;
+    my $next = $matrixdata->{"CD.$ulc.$dr"}->{"Data"};
+    for ($r = 0; $r < $visualr; $r++) {
+        $data = $next;
+        $dr++;
+        $next = $matrixdata->{"CD.$ulc.$dr"}->{"Data"};
+        if (not defined $next or $data eq $next) {
+            for ($c = 0; $c < $visualc; $c++) {
+                $matrixwidget->{"VC.$c.$r"}->{-canvasscr}->attroff(A_UNDERLINE);
+            }
+        }
+        else {
+            for ($c = 0; $c < $visualc; $c++) {
+                $matrixwidget->{"VC.$c.$r"}->{-canvasscr}->attron(A_UNDERLINE);
+            }
+        }
+    }
+
     #   key bindings and event handling
     if (not $cachehit) {
         for ($c = 0; $c < $visualc; $c++) {
             for ($r = 0; $r < $visualr; $r++) {
-                $matrixwidget->{"VC.$c.$r"}->set_binding(\&tabbinder, "\cI", KEY_ENTER(), KEY_F(2));
+                $matrixwidget->{"VC.$c.$r"}->set_binding(\&tabbinder, CUI_TAB(), KEY_ENTER(), KEY_F(2), CUI_ESCAPE());
                 $matrixwidget->{"VC.$c.$r"}->set_event("-onfocus",
                     sub {
                         my $this = shift;
                         my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this}};
                         my ($dc, $dr) = @{$matrixwidgetdatacell->{$this}};
-                        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->standout unless($editcell);
+                        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attron(A_STANDOUT) unless($editcell);
                         if (defined $matrixdata->{"CD.$dc.$dr"}->{"Focuscallback"}) {
                             &{$matrixdata->{"CD.$dc.$dr"}->{"Focuscallback"}}($matrixwidget, $dc, $dr, $vc, $vr);
                         }
@@ -487,7 +510,7 @@ sub layoutmatrixwidget {
                         my $this = shift;
                         my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this}};
                         my ($dc, $dr) = @{$matrixwidgetdatacell->{$this}};
-                        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->standend;
+                        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
                         if ($editcell) {
                             if (defined $matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}) {
                                 $datacoll = &{$matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}}($matrixwidget, $dc, $dr, $vc, $vr);
@@ -504,7 +527,7 @@ sub layoutmatrixwidget {
 sub tabbinder {
     my $this = shift;
     my $key = shift;
-    $key = "\cD" if ($key == KEY_DC()); #make CTRL-D a normal key
+    $key = "\cD" if ($key eq KEY_DC()); #make CTRL-D a normal key
 
     my $nvc = $matrixwidget->{"NumVisualCol"} - 1;
     my $nvr = $matrixwidget->{"NumVisualRow"} - 1;
@@ -514,10 +537,9 @@ sub tabbinder {
     }
 
     my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this}};
+    my ($dc, $dr) = @{$matrixwidgetdatacell->{$this}};
 
-    if    ($editcell and $key eq "\cI") {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this}};
+    if    ($editcell and $key eq CUI_TAB()) {
         my $datacollabel = $matrixdata->{"CH.$dc"}->{"Keyname"};
         if ($datacollabel =~ m/^(date)$/) {
             $this->{-text} = &{$matrixdata->{"completedate"}}($matrixwidget, $dc, $dr, $vc, $vr, $this->{-text});
@@ -571,21 +593,24 @@ sub tabbinder {
                     -vscrollbar => 1,
                     #-onchange   => \&listbox_callback,
                 );
+                $escapeit = 0;
                 $w->set_binding(
                     sub {
                         my $this = shift;
                         my $key = shift;
+                        $escapeit = 1 if ($key eq CUI_ESCAPE());
                         $this->{-has_modal_focus} = 0;
                     },
-                    KEY_ENTER());
+                    KEY_ENTER(), CUI_ESCAPE());
                 $w->draw;
                 $w->modalfocus;
                 $this->parent->delete("accountselection");
-
-                if (defined $w->{-selected}) {
-                    if (defined $matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}) {
-                        $matrixwidget->{"VC.$vc.$vr"}->{-text} = $accounts->[$w->{-selected}];
-                        &{$matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}}($matrixwidget, $dc, $dr, $vc, $vr);
+                unless ($escapeit) {
+                    if (defined $w->{-selected}) {
+                        if (defined $matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}) {
+                            $matrixwidget->{"VC.$vc.$vr"}->{-text} = $accounts->[$w->{-selected}];
+                            &{$matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}}($matrixwidget, $dc, $dr, $vc, $vr);
+                        }
                     }
                 }
                 $editcell = 0;
@@ -593,15 +618,20 @@ sub tabbinder {
             }
         }
     }
-    elsif ($key == KEY_F(2) or $key == KEY_ENTER) {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this}};
+    elsif ($key eq KEY_F(2) or $key eq KEY_ENTER()) {
         if ($editcell) {
             if (defined $matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}) {
                 $datacoll = &{$matrixdata->{"CD.$dc.$dr"}->{"Blurcallback"}}($matrixwidget, $dc, $dr, $vc, $vr);
             }
         }
-        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->standout;
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attron(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        $editcell = 0;
+        return;
+    }
+    elsif ($key eq CUI_ESCAPE()) {
+        $matrixwidget->{"VC.$vc.$vr"}->{-text} = $matrixdata->{"CD.$dc.$dr"}->{"Data"};
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attron(A_STANDOUT);
         $matrixwidget->{"VC.$vc.$vr"}->draw;
         $editcell = 0;
         return;
@@ -614,7 +644,7 @@ sub tabbinder {
 sub keybinder {
     my $this = shift;
     my $key = shift;
-    $key = "\cD" if ($key == KEY_DC()); #make CTRL-D a normal key
+    $key = "\cD" if ($key eq KEY_DC()); #make CTRL-D a normal key
 
     my $nvc = $matrixwidget->{"NumVisualCol"} - 1;
     my $nvr = $matrixwidget->{"NumVisualRow"} - 1;
@@ -647,15 +677,62 @@ sub keybinder {
     }
 
     my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
+    my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
 
-    if    ($key == KEY_UP() or (not $editcell and $key eq "k")) {
+    if    ($key eq CUI_ESCAPE()) {
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        $this->root->focus('menu');
+        return;
+    }
+    elsif ($key eq "s") {
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        $widget->{"dummy"}->focus;
+        {
+            my $wstat = $widget->{'stat'}->width;
+            my $textl = "Accounting System (AS) - save data and continue editing";
+            my $textr = "F1 Help";
+            my $textf = " " x ($wstat -2 -length($textl) -1 -1 -length($textr) - 2) . " ";
+            my $text  = sprintf("  %s %s %s ", $textl, $textf, $textr);
+            $widget->{'info'}->{-text} = $text;
+            $widget->{'info'}->draw;
+        }
+        $breakout = 1;
+        return;
+    }
+    elsif ($key eq "q") {
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        if ($cui->dialog(-message => " Quit without saving? \n" .
+                                     " Changes will be lost ",
+                                       -buttons => 
+                                       [
+                                            { -label => "[ OK ]",     -value => 1 },
+                                            { -label => "[ Cancel ]", -value => 0 }
+                                       ]
+                                        )
+            ) {
+            $widget->{"dummy"}->focus;
+            $breakout = 2;
+        }
+        return;
+    }
+    elsif ($key eq "x" or $key eq "Z") {
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        $widget->{"dummy"}->focus;
+        $breakout = 3;
+        return;
+    }
+    elsif ($key eq KEY_UP() or $key eq "k") {
         $vr--;
         if ($vr < 0) {
             $vrscroll = -1;
             $vr = 0;
         }
     }
-    elsif ($key == KEY_PPAGE()) {
+    elsif ($key eq KEY_PPAGE()) {
         my $j;
         $j = int($matrixwidget->{"NumVisualRow"} / 3 * 1);
         $j = 10 if ( $j < 10 );
@@ -665,14 +742,14 @@ sub keybinder {
             $vr = 0;
         }
     }
-    elsif ($key == KEY_DOWN() or $key eq "j") {
+    elsif ($key eq KEY_DOWN() or $key eq "j") {
         $vr++;
         if ($vr > $nvr) {
             $vrscroll = 1;
             $vr = $nvr;
         }
     }
-    elsif ($key == KEY_NPAGE()) {
+    elsif ($key eq KEY_NPAGE()) {
         my $j;
         $j = int($matrixwidget->{"NumVisualRow"} / 3 * 1);
         $j = 10 if ( $j < 10);
@@ -682,25 +759,25 @@ sub keybinder {
             $vr = $nvr;
         }
     }
-    elsif ($key == KEY_LEFT() or $key eq "h") {
+    elsif ($key eq KEY_LEFT() or $key eq "h") {
         $vc--;
         if ($vc < 0) {
             $vcscroll = -1;
             $vc = 0;
         }
     }
-    elsif ($key == KEY_RIGHT() or $key eq "l") {
+    elsif ($key eq KEY_RIGHT() or $key eq "l") {
         $vc++;
         if ($vc > $nvc) {
             $vcscroll = 1;
             $vc = $nvc
         }
     }
-    elsif ($key == KEY_HOME() && not $editcell) {
+    elsif ($key eq KEY_HOME()) {
         $vr = 0;
         $vc = 0;
     }
-    elsif ($key == KEY_END() && not $editcell) {
+    elsif ($key eq KEY_END()) {
         $vr = $nvr;
         $vc = $nvc;
     }
@@ -738,7 +815,83 @@ sub keybinder {
     elsif ($key eq "\cF") {
         $vrscroll = +$matrixwidget->{"NumVisualRow"};
     }
-    elsif ($key == KEY_F(1)) {
+    elsif ($key eq "w" or $key eq CUI_TAB()) {
+        my $newdc = undef;
+        my $c;
+        unless (defined $newdc) {
+            #   search jump destination from next cell to the end of the columns
+            for ($c = $dc + 1; $c < $matrixdata->{"Columns"}; $c++) {
+                if ($matrixdata->{"CH.$c"}->{"Coljump"}) {
+                    $newdc = $c;
+                    last;
+                }
+            }
+        }
+        unless (defined $newdc) {
+            $vr++;
+            if ($vr > $nvr) {
+                $vrscroll = 1;
+                $vr = $nvr;
+            }
+            #   search jump destination begin of columns to current cell
+            for ($c = 0; $c < $dc; $c++) {
+                if ($matrixdata->{"CH.$c"}->{"Coljump"}) {
+                    $newdc = $c;
+                    last;
+                }
+            }
+        }
+        return unless(defined $newdc); #do not change focus if no jump destination found
+        my $deltac = $newdc - $dc;
+        $vc += $deltac;
+        if ($vc < 0) {
+            $vcscroll = $vc;
+            $vc = 0;
+        }
+        if ($vc > $nvc) {
+            $vcscroll = $vc;
+            $vc = $nvc
+        }
+    }
+    elsif ($key eq "b" or $key eq KEY_BTAB()) {
+        my $newdc = undef;
+        my $c;
+        unless (defined $newdc) {
+            #   search jump destination from previous cell to the begin of the columns
+            for ($c = $dc - 1; $c >= 0; $c--) {
+                if ($matrixdata->{"CH.$c"}->{"Coljump"}) {
+                    $newdc = $c;
+                    last;
+                }
+            }
+        }
+        unless (defined $newdc) {
+            $vr--;
+            if ($vr < 0) {
+                $vrscroll = -1;
+                $vr = 0;
+            }
+            #   search jump destination end of columns to current cell
+            for ($c = $matrixdata->{"Columns"}; $c > $dc; $c--) {
+                if ($matrixdata->{"CH.$c"}->{"Coljump"}) {
+                    $newdc = $c;
+                    last;
+                }
+            }
+        }
+        return unless(defined $newdc); #do not change focus if no jump destination found
+        my $deltac = $newdc - $dc;
+        $vc += $deltac;
+        if ($vc < 0) {
+            $vcscroll = $vc;
+            $vc = 0;
+        }
+        if ($vc > $nvc) {
+            $vcscroll = $vc;
+            $vc = $nvc
+        }
+    }
+    elsif ($key eq KEY_F(1)) {
         my $text = `man as.cui 2>/dev/null` || "Cannot launch man or man page for as.cui not found";
         my $w = $this->parent->add("help", 'TextViewer',
             -x          => $matrixwidget->{"XOffset"},
@@ -756,30 +909,38 @@ sub keybinder {
                 my $key = shift;
                 $this->{-has_modal_focus} = 0;
             },
-            KEY_ENTER());
+            KEY_ENTER(), CUI_ESCAPE());
         $w->draw;
         $w->modalfocus;
         $this->parent->delete("help");
         $datacoll = 1;
     }
-    elsif ($key eq " ") {
+    elsif (($key eq CUI_SPACE()) and $matrixdata->{"CH.$dc"}->{"Coleditable"}) {
         $matrixwidget->{"VC.$vc.$vr"}->{-text} = '';
         $matrixwidget->{"VC.$vc.$vr"}->{-pos} = 0;
-        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->standend;
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
         $matrixwidget->{"VC.$vc.$vr"}->draw;
         $editcell = 1;
         return; #do not change focus
     }
-    elsif ($key == KEY_F(2) or $key == KEY_ENTER) {
-        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->standend;
+    elsif (($key eq "i" or $key eq KEY_F(2) or $key eq KEY_ENTER()) and $matrixdata->{"CH.$dc"}->{"Coleditable"}) {
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
+        $matrixwidget->{"VC.$vc.$vr"}->draw;
+        $editcell = 1;
+        return; #do not change focus
+    }
+    elsif (($key eq "A") and $matrixdata->{"CH.$dc"}->{"Coleditable"}) {
+        my $w = $matrixdata->{"CH.$dc"}->{"Widthmax"};
+        my $p = length($matrixwidget->{"VC.$vc.$vr"}->{-text});
+        $p = $w if ($p > $w);
+        $matrixwidget->{"VC.$vc.$vr"}->{-pos} = $p;
+        $matrixwidget->{"VC.$vc.$vr"}->{-canvasscr}->attroff(A_STANDOUT);
         $matrixwidget->{"VC.$vc.$vr"}->draw;
         $editcell = 1;
         return; #do not change focus
     }
     elsif ($key eq "o") {
         if (defined $matrixdata->{"onanewline"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             &{$matrixdata->{"onanewline"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
@@ -793,8 +954,6 @@ sub keybinder {
     }
     elsif ($key eq "O") {
         if (defined $matrixdata->{"Onanewline"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             &{$matrixdata->{"Onanewline"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
@@ -803,8 +962,6 @@ sub keybinder {
     }
     elsif ($key eq "d") {
         if (defined $matrixdata->{"deletedelete"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             if (&{$matrixdata->{"deletedelete"}}($matrixwidget, $dc, $dr, $vc, $vr)) {
                 # implicit immediate KEY_UP when trailing line was deleted
                 $vr--;
@@ -819,13 +976,56 @@ sub keybinder {
     }
     elsif ($key eq "u") {
         if (defined $matrixdata->{"undo"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             return unless &{$matrixdata->{"undo"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
     }
-    elsif ($key == KEY_F(7)) {
+    elsif ($key eq KEY_F(6)) {
+        my $c;
+        my $values = [];
+        my $labels = {};
+        my $select = {};
+        for ($c = 0; $c < $matrixdata->{"Columns"}; $c++) {
+            push @{$values}, "$c";
+            $labels->{"$c"} = $matrixdata->{"CH.$c"}->{"Label"};
+            $select->{"$c"} = $matrixdata->{"CH.$c"}->{"Coljump"} ? 1 : 0;
+        }
+        my $w = $this->add("coljump", 'Listbox',
+            -x          => $matrixwidget->{"XOffset"},
+            -y          => $matrixwidget->{"YOffset"},
+            -values     => $values,
+            -labels     => $labels,
+            -selected   => $select,
+            -width      => 20,
+            -height     => 20,
+            -border     => 1,
+            -multi      => 1,
+            -title      => 'Jump',
+            -vscrollbar => 1,
+            #-onchange   => \&listbox_callback,
+        );
+        $escapeit = 0;
+        $w->set_binding(
+            sub {
+                my $this = shift;
+                my $key = shift;
+                $escapeit = 1 if ($key eq CUI_ESCAPE());
+                $this->{-has_modal_focus} = 0;
+            },
+            KEY_ENTER(), CUI_ESCAPE());
+        $w->draw;
+        $w->modalfocus;
+        $this->delete("coljump");
+        unless ($escapeit) {
+            for ($c = 0; $c < $matrixdata->{"Columns"}; $c++) {
+                $matrixdata->{"CH.$c"}->{"Coljump"} = $select->{"$c"} ? 1 : 0;
+            }
+        }
+        $this->root->focus(undef, 1);
+        $datacoll = 1;
+        return; #do not change focus
+    }
+    elsif ($key eq KEY_F(7)) {
         my $c;
         my $values = [];
         my $labels = {};
@@ -849,26 +1049,28 @@ sub keybinder {
             -vscrollbar => 1,
             #-onchange   => \&listbox_callback,
         );
+        $escapeit = 0;
         $w->set_binding(
             sub {
                 my $this = shift;
                 my $key = shift;
+                $escapeit = 1 if ($key eq CUI_ESCAPE());
                 $this->{-has_modal_focus} = 0;
             },
-            KEY_ENTER());
+            KEY_ENTER(), CUI_ESCAPE());
         $w->draw;
         $w->modalfocus;
-        #foreach my $key (keys %{$w})      { printf STDERR "       key=$key, value=$w->{$key}\n" };
-        #foreach my $sel (keys %{$select}) { printf STDERR "select sel=$sel, value=$select->{$sel}\n" };
-        for ($c = 0; $c < $matrixdata->{"Columns"}; $c++) {
-            $matrixdata->{"CH.$c"}->{"Colhide"} = $select->{"$c"} ? 0 : 1;
-        }
         $this->delete("hideandseek");
+        unless ($escapeit) {
+            for ($c = 0; $c < $matrixdata->{"Columns"}; $c++) {
+                $matrixdata->{"CH.$c"}->{"Colhide"} = $select->{"$c"} ? 0 : 1;
+            }
+        }
         $this->root->focus(undef, 1);
         $datacoll = 1;
         return; #do not change focus
     }
-    elsif ($key == KEY_F(8)) {
+    elsif ($key eq KEY_F(8)) {
         my $c;
         my $i;
         my $values = [];
@@ -898,9 +1100,6 @@ sub keybinder {
             -values     => $values,
             -selected   => $select,
             -labels     => $labels,
-            #-values     => ["b", "a", "c"],
-            #-selected   => {0 => 1, 2 => 1}, #means 0 (b/bar) and 2 (c/quux) is selected
-            #-labels     => {"a" => "foo", "b" => "bar", "c" => "quux"},
             -width      => 20,
             -height     => 20,
             -border     => 1,
@@ -913,30 +1112,23 @@ sub keybinder {
             sub {
                 my $this = shift;
                 my $key = shift;
+                $escapeit = 1 if ($key eq CUI_ESCAPE());
                 $this->{-has_modal_focus} = 0;
             },
-            KEY_ENTER());
+            KEY_ENTER(), CUI_ESCAPE());
         $w->draw;
         $w->modalfocus;
-
-        #foreach my $key (sort keys %{$w}) { printf STDERR "key=$key, value=$w->{$key}\n" };
-        #printf STDERR "values[] = @{$w->{-values}}\n";
-        #foreach my $key (sort keys %{$w->{-selected}}) { printf STDERR "key=$key, value=$w->{-selected}->{$key} indirect value $w->{-values}->[$key]\n" };
-        #foreach my $key (keys %{$w->{-selected}}) { printf STDERR "$w->{-values}->[$key]\n" if ($w->{-selected}->{$key}); };
-        #die "buh3";
-
-        $matrixdata->{"Sortorder"} = [];
-        for ($i = 0; $i <= $#{$values}; $i++) {
-            $c = $w->{-values}->[$i];
-            #printf STDERR "$c (at index $i) = $w->{-selected}->{$i}\n";
-            push @{$matrixdata->{"Sortorder"}}, $c if ($w->{-selected}->{$i});
-        }
-
         $this->delete("sortselection");
+        unless ($escapeit) {
+            $matrixdata->{"Sortorder"} = [];
+            for ($i = 0; $i <= $#{$values}; $i++) {
+                $c = $w->{-values}->[$i];
+                #printf STDERR "$c (at index $i) = $w->{-selected}->{$i}\n";
+                push @{$matrixdata->{"Sortorder"}}, $c if ($w->{-selected}->{$i});
+            }
+        }
         $this->root->focus(undef, 1);
         if (defined $matrixdata->{"sort"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             &{$matrixdata->{"sort"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
@@ -944,30 +1136,22 @@ sub keybinder {
     }
     elsif ($key eq "y") {
         if (defined $matrixdata->{"yankcell"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             &{$matrixdata->{"yankcell"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
     }
     elsif ($key eq "Y") {
         if (defined $matrixdata->{"Yankline"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             &{$matrixdata->{"Yankline"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
     }
     elsif ($key eq "p") {
         if (defined $matrixdata->{"paste"} and $matrixdata->{"clipboardtype"} eq "cell") {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             return unless &{$matrixdata->{"paste"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
         elsif (defined $matrixdata->{"paste"} and $matrixdata->{"clipboardtype"} eq "line") {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             return unless &{$matrixdata->{"paste"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
             $vr++;
@@ -979,8 +1163,6 @@ sub keybinder {
     }
     elsif ($key eq "P") {
         if (defined $matrixdata->{"Paste"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             return unless &{$matrixdata->{"Paste"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
@@ -989,15 +1171,23 @@ sub keybinder {
     }
     elsif ($key eq ".") {
         if (defined $matrixdata->{"currentdatetime"}) {
-            my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-            my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
             return unless &{$matrixdata->{"currentdatetime"}}($matrixwidget, $dc, $dr, $vc, $vr);
             $datacoll = 1;
         }
     }
-    elsif ($key == KEY_F(9)) {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
+    elsif ($key eq ":") {
+        if (defined $matrixdata->{"copycellfromprevrow"}) {
+            return unless &{$matrixdata->{"copycellfromprevrow"}}($matrixwidget, $dc, $dr, $vc, $vr);
+            $datacoll = 1;
+        }
+    }
+    elsif ($key eq ";") {
+        if (defined $matrixdata->{"copycellfromsuccrow"}) {
+            return unless &{$matrixdata->{"copycellfromsuccrow"}}($matrixwidget, $dc, $dr, $vc, $vr);
+            $datacoll = 1;
+        }
+    }
+    elsif ($key eq KEY_F(9)) {
         my $width = $matrixdata->{"CH.$dc"}->{"Width"};
         if ($width != $matrixdata->{"CH.$dc"}->{"Widthmin"}) {
             $matrixdata->{"CH.$dc"}->{"Width"} = $matrixdata->{"CH.$dc"}->{"Widthmin"};
@@ -1005,9 +1195,7 @@ sub keybinder {
         }
         return; #do not change focus
     }
-    elsif ($key == KEY_F(10)) {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
+    elsif ($key eq KEY_F(10)) {
         my $width = $matrixdata->{"CH.$dc"}->{"Width"};
         $width--;
         if ($width >= $matrixdata->{"CH.$dc"}->{"Widthmin"}) {
@@ -1016,9 +1204,7 @@ sub keybinder {
         }
         return; #do not change focus
     }
-    elsif ($key == KEY_F(11)) {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
+    elsif ($key eq KEY_F(11)) {
         my $width = $matrixdata->{"CH.$dc"}->{"Width"};
         $width++;
         if ($width <= $matrixdata->{"CH.$dc"}->{"Widthmax"}) {
@@ -1027,9 +1213,7 @@ sub keybinder {
         }
         return; #do not change focus
     }
-    elsif ($key == KEY_F(12)) {
-        my ($vc, $vr) = @{$matrixwidgetvisualcell->{$this->getfocusobj}};
-        my ($dc, $dr) = @{$matrixwidgetdatacell->{$this->getfocusobj}};
+    elsif ($key eq KEY_F(12)) {
         my $width = $matrixdata->{"CH.$dc"}->{"Width"};
         if ($width != $matrixdata->{"CH.$dc"}->{"Widthmax"}) {
             $matrixdata->{"CH.$dc"}->{"Width"} = $matrixdata->{"CH.$dc"}->{"Widthmax"};
@@ -1037,6 +1221,9 @@ sub keybinder {
         }
         return; #do not change focus
     }
+    else {
+        $this->dobeep;
+    }
     $matrixwidget->{"Cursorcol"} = $vc;
     $matrixwidget->{"Cursorrow"} = $vr;
     if ($vcscroll == 0 && $vrscroll == 0) {
@@ -1086,17 +1273,14 @@ sub drawfocusmatrixwidget {
     if ($lastc >= 0 and $lastr >= 0) {
         $c = $lastc if ($c > $lastc);
         $r = $lastr if ($r > $lastr);
-        $matrixwidget->{"VC.$c.$r"}->{-canvasscr}->standout;
+        $matrixwidget->{"VC.$c.$r"}->{-canvasscr}->attron(A_STANDOUT);
         $matrixwidget->{"VC.$c.$r"}->draw;
         $matrixwidget->{"VC.$c.$r"}->focus;
     }
 }
 
-sub ascui {
-    ($matrixdata, $verbose) = (@_);
-
-    #   in CUI mode we do not want STDERR to clutter our screen.
-    ($verbose && open STDERR, ">>2.log") || (open STDERR, ">/dev/null");
+sub ascuinew {
+     ($matrixdata, $verbose) = (@_);
 
     #   create the root object
     $cui = new Curses::UI (
@@ -1149,7 +1333,7 @@ sub ascui {
     #   add a status line
     my $wstat = $widget->{'stat'}->width;
     my $textl = "Accounting System (AS)";
-    my $textr = "CTRL+X: menu  CTRL+Q: quit";
+    my $textr = "F1 Help";
     my $textf = " " x ($wstat -2 -length($textl) -1 -1 -length($textr) - 2) . " ";
     my $text  = sprintf("  %s %s %s ", $textl, $textf, $textr);
     $widget->{'status'}->{-text} = $text;
@@ -1170,11 +1354,20 @@ sub ascui {
         -height => 1
     );
 
-    $cui->set_binding(sub{ $breakout = 1 }, "\cQ");
-    $cui->set_binding(sub{ shift()->root->focus('menu') }, "\cX");
     $widget->{'container'}->set_binding(\&keybinder, ''); #now we catch all chars not just @keys_move
     $cui->draw;
+}
 
+sub ascuido {
+    {
+        my $wstat = $widget->{'stat'}->width;
+        my $textl = "Accounting System (AS)";
+        my $textr = "F1 Help";
+        my $textf = " " x ($wstat -2 -length($textl) -1 -1 -length($textr) - 2) . " ";
+        my $text  = sprintf("  %s %s %s ", $textl, $textf, $textr);
+        $widget->{'info'}->{-text} = $text;
+        $widget->{'info'}->draw;
+    }
     &newmatrixwidget();
     my ($upperleftcol, $cursorcol) = ($matrixdata->{"CK.date"}, 0); #&rightmostvisiblefit();
     my ($upperleftrow, $cursorrow) = &bottommostvisiblefit();
@@ -1230,7 +1423,10 @@ sub ascui {
     } until ($breakout);
     &destroymatrixwidget();
 
-    # $cui->dialog("Bye bye!");
+    return "save" if ($breakout == 1);
+    return "quit" if ($breakout == 2);
+    return "exit" if ($breakout == 3);
+    return;
 }
 
 1;
